# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import final

from gravitino.dto.rel.partitioning.partitioning import SingleFieldPartitioning


@final
class HourPartitioningDTO(SingleFieldPartitioning):
    """Represents a HourPartitioning Data Transfer Object (DTO)."""

    def __init__(self, /, *field_name: str):
        super().__init__(list(field_name))

    def strategy(self) -> SingleFieldPartitioning.Strategy:
        """Returns the strategy of the partitioning.

        Returns:
            Strategy: The strategy of the partitioning.
        """
        return SingleFieldPartitioning.Strategy.HOUR
