# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from gravitino.api.job.job_handle import JobHandle
from gravitino.dto.job.job_dto import JobDTO


class GenericJobHandle(JobHandle):
    """Represents a generic job handle."""

    def __init__(self, job_dto: JobDTO):
        self._job_dto = job_dto

    def job_template_name(self) -> str:
        return self._job_dto.job_template_name()

    def job_id(self) -> str:
        return self._job_dto.job_id()

    def job_status(self):
        return self._job_dto.status()
